@@ -8,9 +8,9 @@ module ActiveSupport
 
   module VERSION
     MAJOR = 6
-    MINOR = 0
-    TINY  = 3
-    PRE   = "4"
+    MINOR = 1
+    TINY  = 0
+    PRE   = nil
 
     STRING = [MAJOR, MINOR, TINY, PRE].compact.join(".")
   end
